﻿__device__ int find_pos(double sca, const double *grid, const int N) {

    int pos = 0;
    int i1 = 0;
    while (i1 < N) {
        if (sca >= grid[i1]) {
            pos = i1;
            i1 = i1 + 1;
        }
        else {
            i1 = N;
        }
    }
    return pos;
}

__device__ double lininter(double sca, const double *y, const double *grid, const int N) {
    int pos = find_pos(sca, grid, N);

    //if (pos == 0) {
    //    pos = 1;
    //}
    if (pos == N-1) {
        pos = N - 2;
    }

    double y0 = y[pos] + (y[pos + 1] - y[pos]) * (sca - grid[pos]) / (grid[pos + 1] - grid[pos]);

    return y0;

}

__global__ void golden(double *Vnew, double *policy, const double *V, const double *P, const double *agrid, double *EV, const double *coh, const double *par, const int* index)
{
    //Computes one update of the value function
    int COL = blockIdx.x * blockDim.x + threadIdx.x;
    int ROW = blockIdx.y * blockDim.y + threadIdx.y;

    //read out parameter vector
    double amin = par[0];
    double beta = par[1];

    int Na = index[0];
    int Nz = index[1];

    //initialization
    double r1 = (sqrt(5.0) - 1.0) / 2.0;
    double r2 = 1.0 - r1;
    double lb;
    double ub;
    double util;
    double cont;
    double val_b;
    double val_c;
    double bb;
    double cc;
    int pos;
    

    if (ROW < Na && COL < Nz) {
        lb = amin;
        double resources = coh[ROW + COL * Na];
        ub = resources;

        bb = r1 * lb + r2 * ub;
        //cont = lininter(bb, (V+ COL * Na), agrid, Na); //pass the correct portion of V
        pos = find_pos(bb, agrid, Na); //choice index
        if (pos == Na - 1) {
            pos = Na - 2;
        }
        //check if the expected value function has already been computed. If not, compute it
        if (EV[pos+ COL * Na] < -99999) { //as na is small enough, I could read, for each COL, EV into shared memory. probably fastest to fill it up at the beginning for each pos
            double EVE = 0;   //I acess agrid often, probably should be in shared memory
            for (int i2 = 0; i2 < Nz; i2++) {
                EVE += P[COL * Nz + i2] * V[pos + i2 * Na];
            }
            EV[pos + COL * Na] = EVE;
        }
        if (EV[pos+1 + COL * Na] < -99999) {
            double EVE = 0;
            for (int i2 = 0; i2 < Nz; i2++) {
                EVE += P[COL * Nz + i2] * V[pos+1 + i2 * Na];
            }
            EV[pos+1 + COL * Na] = EVE;
        }
        util = log(resources - bb);
        cont = EV[pos + COL * Na] + (EV[pos + 1 + COL * Na] - EV[pos + COL * Na]) * (bb - agrid[pos]) / (agrid[pos + 1] - agrid[pos]);
        val_b = util + beta * cont;


        cc = r2 * lb + r1 * ub;
        pos = find_pos(cc, agrid, Na); //choice index
        if (pos == Na - 1) {
            pos = Na - 2;
        }
        //check if the expected value function has already been computed. If not, compute it
        if (EV[pos + COL * Na] < -99999) {
            double EVE = 0;
            for (int i2 = 0; i2 < Nz; i2++) {
                EVE += P[COL * Nz + i2] * V[pos + i2 * Na];
            }
            EV[pos + COL * Na] = EVE;
        }
        if (EV[pos + 1 + COL * Na] < -99999) {
            double EVE = 0;
            for (int i2 = 0; i2 < Nz; i2++) {
                EVE += P[COL * Nz + i2] * V[pos + 1 + i2 * Na];
            }
            EV[pos + 1 + COL * Na] = EVE;
        }
        util = log(resources - cc);
        cont = EV[pos + COL * Na] + (EV[pos + 1 + COL * Na] - EV[pos + COL * Na]) * (cc - agrid[pos]) / (agrid[pos + 1] - agrid[pos]);
        val_c = util + beta * cont;

        
        while (abs(lb - ub) > 0.001 * (abs(bb) + abs(cc))) {
            if (val_b > val_c) {
                ub = cc;
                cc = bb;
                val_c = val_b;
                bb = r1 * lb + r2 * ub;
                
                bb = r1 * lb + r2 * ub;
                pos = find_pos(bb, agrid, Na); //choice index
                if (pos == Na - 1) {
                    pos = Na - 2;
                }
                //check if the expected value function has already been computed. If not, compute it
                if (EV[pos + COL * Na] < -99999) {
                    double EVE = 0;
                    for (int i2 = 0; i2 < Nz; i2++) {
                        EVE += P[COL * Nz + i2] * V[pos + i2 * Na];
                    }
                    EV[pos + COL * Na] = EVE;
                }
                if (EV[pos + 1 + COL * Na] < -99999) {
                    double EVE = 0;
                    for (int i2 = 0; i2 < Nz; i2++) {
                        EVE += P[COL * Nz + i2] * V[pos + 1 + i2 * Na];
                    }
                    EV[pos + 1 + COL * Na] = EVE;
                }
                util = log(resources - bb);
                cont = EV[pos + COL * Na] + (EV[pos + 1 + COL * Na] - EV[pos + COL * Na]) * (bb - agrid[pos]) / (agrid[pos + 1] - agrid[pos]);
                val_b = util + beta * cont;

            }
            else {
                lb = bb;
                bb = cc;
                val_b = val_c;
                cc = r2 * lb + r1 * ub;
                pos = find_pos(cc, agrid, Na); //choice index
                if (pos == Na - 1) {
                    pos = Na - 2;
                }
                //check if the expected value function has already been computed. If not, compute it
                if (EV[pos + COL * Na] < -99999) {
                    double EVE = 0;
                    for (int i2 = 0; i2 < Nz; i2++) {
                        EVE += P[COL * Nz + i2] * V[pos + i2 * Na];
                    }
                    EV[pos + COL * Na] = EVE;
                }
                if (EV[pos + 1 + COL * Na] < -99999) {
                    double EVE = 0;
                    for (int i2 = 0; i2 < Nz; i2++) {
                        EVE += P[COL * Nz + i2] * V[pos + 1 + i2 * Na];
                    }
                    EV[pos + 1 + COL * Na] = EVE;
                }
                util = log(resources - cc);
                cont = EV[pos + COL * Na] + (EV[pos + 1 + COL * Na] - EV[pos + COL * Na]) * (cc - agrid[pos]) / (agrid[pos + 1] - agrid[pos]);
                val_c = util + beta * cont;
            }

        }

        if (val_b > val_c) {
            policy[ROW + COL * Na] = bb;
            Vnew[ROW + COL * Na] = val_b;
        }
        else {
            policy[ROW + COL * Na] = cc;
            Vnew[ROW + COL * Na] = val_c;
        }
        
    }


}