clear all
close all
clc

%household problem with Golden-Section-Search

%% Parameters

%preferences
beta = 0.96; %time preferences
gamma = 2;     %risk aversion
gamma1 = 1-gamma;



% Idiosyncratic productivity Process
nz = 30;
rho_e = 0.95;
sigma_e = 0.045;
[epsilon,PI_e] = Tauchen(rho_e,nz,sigma_e^2/(1-rho_e^2),0);
epsilon = exp(epsilon)';


%capital grid
na = 128; 
amin = 0;
amax = 5*max(epsilon);
agrid = linspace(amin,amax,na)';


%guess interest rate
r = 0.035/4; %interest rate flexible asset

coh = repmat((1+r)*agrid,[1 nz])+repmat(epsilon',[na 1]);

V_old = 1/(1-beta)*log(repmat(r*agrid,[1 nz])+repmat(epsilon',[na 1])); %guess value function
EV_old = -999999*ones(na,nz);

%% value function itteration CPU MATLAB

PIe_m = PI_e';
PIe_m = PIe_m(:);
agrid_m = agrid';
coh_m = coh(:)';
V_m = V_old(:)';
EV_m = EV_old(:)';
Vnew_m = zeros(1,na*nz);
policy_m = zeros(1,na*nz);

tic
crit = 1;
while crit>0.0001
    for j = 1:nz
        for i = 1:na
            [Vnew_m(i + (j-1) * na),policy_m(i + (j-1) * na)] = golden_mat(V_m,agrid_m,PIe_m,EV_m,coh_m,beta,amin,na,nz,i,j);
        end
    end
    crit = max(max(abs(Vnew_m-V_m)));
    V_m = Vnew_m;
end
done = 1;
toc
Vnew_m = reshape(Vnew_m,[na nz]);

%% value function itteration CPU with Matlab's fminbnd

Vnew = zeros(na,nz);
aprime = zeros(na,nz);
V = V_old;
tic
crit = 1;
while crit>0.0001
    %expected value function
    EV = zeros(na,nz);
    for i = 1:nz %state
        for j = 1:nz %transition to
            EV(:,i) = EV(:,i)+V(:,j)*PI_e(i,j);
        end
    end
    
    %solve maximization problem   V = max\{log(c(a'))+beta*EV(a',z)\}
    for i = 1:na %asset state
        for j = 1:nz %productivity state
            coh_aux = coh(i,j); %maximum consumption (asset choice)
%             [Vnew(i,j),aprime(i,j)] = golden_mex(beta*EV(:,j),agrid,amin,coh,1,int32(na));
            [aprime(i,j),Vnew(i,j)] = fminbnd(@(x) value(x,EV(:,j),coh_aux,beta,agrid),amin,coh_aux); %find minimum on [amin,coh]
        end
    end
    Vnew=-Vnew;
    crit = max(max(abs(Vnew-V)));
    V = Vnew;
    
end
done = 1;
toc

%% value function itteration GPU 

dims = size(V_old);
k = parallel.gpu.CUDAKernel('golden.ptx','golden.cu');
k.ThreadBlockSize = [1 dims(1) 1]; %number of threads
k.GridSize = [dims(2) 1 1]; %number of blocks

V_new_d = gpuArray(zeros(1,dims(1)*dims(2)));
policy_a_d = gpuArray(zeros(1,dims(1)*dims(2)));

PIe_d = PI_e';
PIe_d = gpuArray(PIe_d(:));
agrid_d = gpuArray(agrid');
epsilon_d = gpuArray(epsilon');
coh_d = gpuArray(coh(:)');
par_d = gpuArray([amin beta]);
V_d = gpuArray(V_old(:)');
EV_d = gpuArray(EV_old(:)');
index_d = gpuArray(int32([dims(1) dims(2)]));



tic
crit=1;
while crit>0.0001
    
    [V_new_d,policy_a_d] = feval(k,V_new_d,policy_a_d,V_d,PIe_d,agrid_d,EV_d,coh_d,par_d,index_d);
    


    crit = gather(max(max(max(abs(V_new_d-V_d)))));
    V_d = V_new_d;
end
done = 1;
toc

Vnew_GPU = gather(V_new_d);
Vnew_GPU = reshape(Vnew_GPU,[dims(1),dims(2)]);


%% value function itteration GPU shared memory

dims = size(V_old);
k = parallel.gpu.CUDAKernel('golden_share.ptx','golden_share.cu');
% k = parallel.gpu.CUDAKernel('comp_EVE.ptx','comp_EVE.cu');
k.ThreadBlockSize = [1 dims(1) 1]; %number of threads
k.GridSize = [dims(2) 1 1]; %number of blocks

V_new_d = gpuArray(zeros(1,dims(1)*dims(2)));
policy_a_d = gpuArray(zeros(1,dims(1)*dims(2)));

PIe_d = PI_e';
PIe_d = gpuArray(PIe_d(:));
agrid_d = gpuArray(agrid');
epsilon_d = gpuArray(epsilon');
coh_d = gpuArray(coh(:)');
par_d = gpuArray([amin beta]);
V_d = gpuArray(V_old(:)');
EV_d = gpuArray(EV_old(:)');
index_d = gpuArray(int32([dims(1) dims(2)]));



tic
crit=1;
while crit>0.0001
    
    [V_new_d,policy_a_d] = feval(k,V_new_d,policy_a_d,V_d,PIe_d,agrid_d,coh_d,par_d,index_d);


    crit = gather(max(max(max(abs(V_new_d-V_d)))));
    V_d = V_new_d;
end
done = 1;
toc

Vnew_GPU2 = gather(V_new_d);
Vnew_GPU2 = reshape(Vnew_GPU2,[dims(1),dims(2)]);

policy_a_GPU = gather(policy_a_d);
policy_a_GPU = reshape(policy_a_GPU,[dims(1),dims(2)]);

